@@ -24,14 +24,14 @@
 
 import itertools
 import math
-from typing import Callable, Optional
+from typing import Callable, Optional, Text, Union
 
 import numpy as np
 import torch
 from einops import rearrange
 from pyannote.core import Annotation, SlidingWindow, SlidingWindowFeature
 from pyannote.metrics.diarization import GreedyDiarizationErrorRate
-from pyannote.pipeline.parameter import Uniform
+from pyannote.pipeline.parameter import ParamDict, Uniform
 
 from pyannote.audio import Audio, Inference, Model, Pipeline
 from pyannote.audio.core.io import AudioFile
@@ -61,6 +61,10 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
     segmentation : Model, str, or dict, optional
         Pretrained segmentation model. Defaults to "pyannote/segmentation@2022.07".
         See pyannote.audio.pipelines.utils.get_model for supported format.
+    segmentation_duration: float, optional
+        The segmentation model is applied on a window sliding over the whole audio file.
+        `segmentation_duration` controls the duration of this window. Defaults to the
+        duration used when training the model (model.specifications.duration).
     segmentation_step: float, optional
         The segmentation model is applied on a window sliding over the whole audio file.
         `segmentation_step` controls the step of this window, provided as a ratio of its
@@ -78,6 +82,14 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         Batch size used for speaker segmentation. Defaults to 32.
     embedding_batch_size : int, optional
         Batch size used for speaker embedding. Defaults to 32.
+    der_variant : dict, optional
+        Optimize for a variant of diarization error rate.
+        Defaults to {"collar": 0.0, "skip_overlap": False}. This is used in `get_metric`
+        when instantiating the metric: GreedyDiarizationErrorRate(**der_variant).
+    use_auth_token : str, optional
+        When loading private huggingface.co models, set `use_auth_token`
+        to True or to a string containing your hugginface.co authentication
+        token that can be obtained by running `huggingface-cli login`
 
     Usage
     -----
@@ -86,23 +98,36 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
     >>> diarization = pipeline("/path/to/audio.wav", num_speakers=4)
     >>> diarization = pipeline("/path/to/audio.wav", min_speakers=2, max_speakers=10)
 
+    Hyper-parameters
+    ----------------
+    segmentation.threshold
+    segmentation.min_duration_off
+    clustering.???
     """
 
     def __init__(
         self,
         segmentation: PipelineModel = "pyannote/segmentation@2022.07",
+        segmentation_duration: float = None,
         segmentation_step: float = 0.1,
         embedding: PipelineModel = "speechbrain/spkrec-ecapa-voxceleb@5c0be3875fda05e81f3c004ed8c7c06be308de1e",
         embedding_exclude_overlap: bool = False,
         clustering: str = "HiddenMarkovModelClustering",
         embedding_batch_size: int = 32,
         segmentation_batch_size: int = 32,
+        der_variant: dict = None,
+        use_auth_token: Union[Text, None] = None,
     ):
 
         super().__init__()
 
-        self.segmentation = segmentation
+        self.segmentation_model = segmentation
+        model: Model = get_model(segmentation, use_auth_token=use_auth_token)
+
         self.segmentation_batch_size = segmentation_batch_size
+        self.segmentation_duration = (
+            segmentation_duration or model.specifications.duration
+        )
         self.segmentation_step = segmentation_step
 
         self.embedding = embedding
@@ -111,27 +136,32 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
 
         self.klustering = clustering
 
+        self.der_variant = der_variant or {"collar": 0.0, "skip_overlap": False}
+
         seg_device, emb_device = get_devices(needs=2)
 
-        model: Model = get_model(segmentation)
         model.to(seg_device)
 
         self._segmentation = Inference(
             model,
-            duration=model.specifications.duration,
-            step=self.segmentation_step * model.specifications.duration,
+            duration=self.segmentation_duration,
+            step=self.segmentation_step * self.segmentation_duration,
             skip_aggregation=True,
             batch_size=self.segmentation_batch_size,
         )
         self._frames: SlidingWindow = self._segmentation.model.introspection.frames
-        self.segmentation_onset = Uniform(0.1, 0.9)
+
+        self.segmentation = ParamDict(
+            threshold=Uniform(0.1, 0.9),
+            min_duration_off=Uniform(0.0, 1.0),
+        )
 
         if self.klustering == "OracleClustering":
             metric = "not_applicable"
 
         else:
             self._embedding = PretrainedSpeakerEmbedding(
-                self.embedding, device=emb_device
+                self.embedding, device=emb_device, use_auth_token=use_auth_token
             )
             self._audio = Audio(sample_rate=self._embedding.sample_rate, mono=True)
             metric = self._embedding.metric
@@ -147,7 +177,8 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
     def default_parameters(self):
 
         if (
-            self.segmentation == "pyannote/segmentation@2022.07"
+            self.segmentation_model == "pyannote/segmentation@2022.07"
+            and self.segmentation_duration == 5.0
             and self.segmentation_step == 0.1
             and self.embedding
             == "speechbrain/spkrec-ecapa-voxceleb@5c0be3875fda05e81f3c004ed8c7c06be308de1e"
@@ -155,7 +186,10 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
             and self.clustering == "HiddenMarkovModelClustering"
         ):
             return {
-                "segmentation_onset": 0.58,
+                "segmentation": {
+                    "threshold": 0.58,
+                    "min_duration_off": 0.0,
+                },
                 "clustering": {
                     "single_cluster_detection": {
                         "quantile": 0.05,
@@ -231,7 +265,7 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
             # "segmentation_onset" hyperparameter and "embedding_exclude_overlap" parameter.
             cache = file.get("training_cache/embeddings", dict())
             if (
-                cache.get("segmentation_onset", None) == self.segmentation_onset
+                cache.get("segmentation.threshold", None) == self.segmentation.threshold
                 and cache.get("embedding_exclude_overlap", None)
                 == self.embedding_exclude_overlap
             ):
@@ -327,7 +361,7 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         # (see comments at the top of this method for more details)
         if self.training:
             file["training_cache/embeddings"] = {
-                "segmentation_onset": self.segmentation_onset,
+                "segmentation.threshold": self.segmentation.threshold,
                 "embedding_exclude_overlap": self.embedding_exclude_overlap,
                 "embeddings": embeddings,
             }
@@ -431,7 +465,7 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         # estimate frame-level number of instantaneous speakers
         count = self.speaker_count(
             segmentations,
-            onset=self.segmentation_onset,
+            onset=self.segmentation.threshold,
             frames=self._frames,
         )
         hook("speaker_counting", count)
@@ -441,7 +475,7 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         # binarize segmentation
         binarized_segmentations: SlidingWindowFeature = binarize(
             segmentations,
-            onset=self.segmentation_onset,
+            onset=self.segmentation.threshold,
             initial_state=False,
         )
 
@@ -486,14 +520,14 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         diarization = self.to_annotation(
             discrete_diarization,
             min_duration_on=0.0,
-            min_duration_off=0.0,
+            min_duration_off=self.segmentation.min_duration_off,
         )
         diarization.uri = file["uri"]
 
         # when reference is available, use it to map hypothesized speakers
         # to reference speakers (this makes later error analysis easier
         # but does not modify the actual output of the diarization pipeline)
-        if "annotation" in file:
+        if "annotation" in file and file["annotation"]:
             return self.optimal_mapping(file["annotation"], diarization)
 
         # when reference is not available, rename hypothesized speakers
@@ -506,4 +540,4 @@ class SpeakerDiarization(SpeakerDiarizationMixin, Pipeline):
         )
 
     def get_metric(self) -> GreedyDiarizationErrorRate:
-        return GreedyDiarizationErrorRate(collar=0.0, skip_overlap=False)
+        return GreedyDiarizationErrorRate(**self.der_variant)
